#include "ObjectRelationalAccess/ObjectRelationalMappingSchema.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "ObjectRelationalAccess/ObjectRelationalNames.h"
#include "RelationalAccess/IColumn.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/SchemaException.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/MessageStream.h"

pool::ObjectRelationalMappingSchema::ObjectRelationalMappingSchema( coral::ISchema& schema ):
  m_schema( schema )
{}


pool::ObjectRelationalMappingSchema::~ObjectRelationalMappingSchema()
{
}

bool
pool::ObjectRelationalMappingSchema::upgradeTables() {
  bool ret = false;
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  try {
     if(!existTables()){
       log << coral::Error << " Mapping Tables do not exist in the schema."<< coral::MessageStream::endmsg;
       return false;
     }
    
     // check if the existing mapping version table has got all of the expected columns
     coral::ITable& versionTable = m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() );
     const coral::ITableDescription& versionTableDescr = versionTable.description();
     try {
       versionTableDescr.columnDescription( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable() );
       log << coral::Warning << " Table \""<< pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() <<"\" is up to date."<< coral::MessageStream::endmsg;
     } catch (const coral::InvalidColumnNameException&){
       ret = true;
       versionTable.schemaEditor().insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable(),
                                                coral::AttributeSpecification::typeNameForType<std::string>());
       log << coral::Warning << " Added column \""<< pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable() << "\" to table \""<< pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() <<"\"."<< coral::MessageStream::endmsg;
       std::ostringstream setClause;
       setClause << pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable() << "= 'POOL_CONTAINER'";
       versionTable.dataEditor().updateRows( setClause.str(),std::string(""),coral::AttributeList());
       versionTable.schemaEditor().setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable() );
     }
  } catch (const coral::Exception& e){
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingSchema::upgradeSchema" );    
  }
  return ret;
}

bool
pool::ObjectRelationalMappingSchema::existTables(){
  bool ret = false;
  std::set<std::string> tableList = listTables();
  size_t nmissing = 0;
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  try {
    for(std::set<std::string>::const_iterator iT=tableList.begin(); iT!=tableList.end();++iT){
      if(!m_schema.existsTable( *iT )){
        nmissing++;
      }
    }
    if(nmissing==0) {
      ret = true;
    } else if(nmissing!=tableList.size()){
      throw pool::ObjectRelationalException( " Mapping Database is corrupted.",
                                             "ObjectRelationalMappingSchema::existsMappingDatabase" );
    }    
  } catch (const coral::Exception& e){
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingSchema::existsMappingDatabase" );
  }
  return ret;
}

void
pool::ObjectRelationalMappingSchema::createTables()
{
 // The versions table.
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  log << coral::Warning << "Creating table \"" << pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() << "\"" << coral::MessageStream::endmsg;
  coral::TableDescription  description0( "ObjectRelationalAccess" );
  description0.setName(pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable());
  description0.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>());
  description0.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable() );
  description0.setUniqueConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable() );
  description0.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>());
  description0.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable() );
  m_schema.createTable( description0 ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
  

 // The class versions table.
  log << coral::Warning << "Creating table \"" << pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable() << "\"" << coral::MessageStream::endmsg;
  coral::TableDescription description1( "ObjectRelationalAccess" );
  description1.setName(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable());
  description1.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>(), 1000, false);
  description1.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable() );
  description1.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>(), 1000, false);
  description1.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable() );
  std::vector<std::string> cols;
  cols.push_back(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable());
  cols.push_back(pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable());
  description1.setUniqueConstraint( cols );
  description1.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>());
  description1.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable() );  
  description1.createForeignKey(
  pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable()+"_FK",
  pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable(),
  pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable(),
  pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable());
  m_schema.createTable( description1 ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
  
  // The table holding the mapping elements.
  log << coral::Warning << "Creating table \"" << pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable() << "\"" <<
         coral::MessageStream::endmsg;
  coral::TableDescription description2( "ObjectRelationalAccess" );
  description2.setName(pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable());
  description2.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::elementTypeFieldInMappingElementTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>() );
  description2.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>() );
  description2.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>() );
  description2.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::variableTypeFieldInMappingElementTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>() );
  description2.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>() );
  description2.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>() );
  description2.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::elementTypeFieldInMappingElementTable() );
  description2.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable() );
  description2.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable() );
  description2.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::variableTypeFieldInMappingElementTable() );
  description2.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() );
  description2.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable() );
  m_schema.createTable( description2 ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
  
  // The table holding the columns in the mapping elements.
  log << coral::Warning << "Creating table \"" << pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementColumnsTable() << "\"" << coral::MessageStream::endmsg;
  coral::TableDescription description3( "ObjectRelationalAccess" );
  description3.setName(pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementColumnsTable());
  description3.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementColumnsTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>() );
  description3.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementColumnsTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>() );
  description3.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementColumnsTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>() );
  description3.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::columnFieldInMappingElementColumnsTable(),
                             coral::AttributeSpecification::typeNameForType<std::string>() );
  description3.insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::positionFieldInMappingElementColumnsTable(),
                             coral::AttributeSpecification::typeNameForType<int>() );
  description3.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementColumnsTable() );
  description3.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementColumnsTable() );
  description3.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementColumnsTable() );
  description3.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::columnFieldInMappingElementColumnsTable() );
  description3.setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::positionFieldInMappingElementColumnsTable() );
  m_schema.createTable( description3 ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
}

std::set<std::string>
pool::ObjectRelationalMappingSchema::listTables(){
  std::set<std::string> tableList;
  tableList.insert( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() );
  tableList.insert( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable() );
  tableList.insert( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable() );
  tableList.insert( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementColumnsTable() );
  return tableList;
}
